"""A module containing the Service Principal credential class"""
from azol.credentials.entraid_credential import EntraIdCredential
from azol.constants import OAUTHFLOWS
from cryptography.hazmat.primitives.serialization import pkcs12
import base64
import os
import logging
import requests

class ADOWorkloadFederationCredential( EntraIdCredential ):
    """
        A credential to log in as a service principal

        IMPORTANT: two dependencies must be met in order to use this credential in a devops pipeline.

                1)
                    This access token must be in the local environment variables in the azure devops
                    pipeline execution environment. to ensure that this is the case, request the
                    system access token in the environment variables of the azure devops job. for example,
                    in the yaml pipeline file, use the following script task to set the system access token
                    to the variable "SYSTEM_ACCESSTOKEN":

                        - script |
                            env
                          env:
                            SYSTEM_ACCESSTOKEN: $(System.AccessToken)

                2) 
                    The system access token will not be able to get an oidc token for a service
                    endpoint unless that service endpoint is used in another task within the same
                    job. The easiest way to make this work is to use the Azure CLI task, and run
                    azol within that task. For example:


                        - task: AzureCLI@2
                            inputs:
                                azureSubscription: 'your-service-connection-name'
                                scriptType: 'bash'
                                scriptLocation: 'InlineScript'
                                inlineScript: |
                                  pip install azol
                                  python -c "
                                  from azol import *;
                                  cred=ADOWorkloadFederationCredential(
                                    client_id='your-client-id', 
                                    service_endpoint_id='your-service-endpoint-id' );
                                  client=ArmClient(cred=cred, tenant='yourtenant.com');
                                  client.fetch_token();
                                  token=client.get_current_token()"
                          env:
                            SYSTEM_ACCESSTOKEN: $(System.AccessToken)
    """

    supportedOAuthFlows = [ OAUTHFLOWS.CLIENT_CREDENTIALS ]
    credentialType="app"
    default_oauth_flow=OAUTHFLOWS.CLIENT_CREDENTIALS
    def __init__( self, client_id, service_endpoint_id, *args, **kwargs ):
        super().__init__( *args, **kwargs)
        self._client_id=client_id
        self._credential_type="ado_oidc"
        self._service_endpoint_id = service_endpoint_id

    def get_credential_type( self ):
        """
            Returns: string - secret or x509, depending on what type of secret was defined for
                     the service principal
        """
        return self._credential_type

    def get_service_endpoint( self ):
        """
            Get the service endpoint id used for this cert

            Returns: azure devops service endpoint id
        """
        return self._service_endpoint_id
    
    def get_system_access_token( self, var_name="SYSTEM_ACCESSTOKEN" ):
        """
            Get the system access token for the workload federation credential. 

            IMPORTANT: A dependency must be met in order to use this credential in a devops pipeline.

                    This access token must be in the local environment variables in the azure devops
                    pipeline execution environment. to ensure that this is the case, request the
                    system access token in the environment variables of the azure devops job. for example,
                    in the yaml pipeline file, use the following script task to set the system access token
                    to the variable "SYSTEM_ACCESSTOKEN":

                        - script |
                            env
                          env:
                            SYSTEM_ACCESSTOKEN: $(System.AccessToken)

            Args:
                - var_name: The environment variable name that contains the system access token.
                            Must match the variable set in the yaml pipeline

            Returns: The system access token for the Azure DevOps job
        """
        try:
            system_access_token=os.environ.get(var_name)
        except Exception as msg:
            logging.error("could not get system access token")
        
        return system_access_token

    def get_oidc_url( self ):
        """
            Gets the OIDC url from the azure devops execution environment. This
            is saved in an environment variable

            Returns: The OIDC URL for Azure Devops
        
        """
        return os.environ.get("SYSTEM_OIDCREQUESTURI")


import ctypes
import ctypes.wintypes
import os
import json

def decrypt_dpapi(encrypted_data):

    buffer = ctypes.create_string_buffer(len(encrypted_data))
    buffer[:len(encrypted_data)] = encrypted_data

    # Call CryptUnprotectData
    class DATA_BLOB(ctypes.Structure):
        _fields_ = [("cbData", ctypes.wintypes.DWORD),
                    ("pbData", ctypes.POINTER(ctypes.c_char))]

    encrypted_blob = DATA_BLOB(len(encrypted_data), buffer)
    decrypted_blob = DATA_BLOB()

    result = ctypes.windll.crypt32.CryptUnprotectData(
        ctypes.byref(encrypted_blob),
        None,
        None,
        None,
        None,
        0,
        ctypes.byref(decrypted_blob)
    )

    if not result:
        raise ctypes.WinError()

    # Extract the decrypted data
    decrypted_data = ctypes.string_at(decrypted_blob.pbData, decrypted_blob.cbData)
    ctypes.windll.kernel32.LocalFree(decrypted_blob.pbData)
    return decrypted_data

def get_azure_cli_credential_file_contents():
    os_name = os.name
    home_directory = os.path.expanduser ("~")
    azure_directory = home_directory + "/.azure"
    msal_file = azure_directory + "/msal_token_cache.bin"
    #throw exception if directory doesnt exist
    if not os.path.isdir(azure_directory):
        raise Exception("Could not find an Azure CLI directory in the user's home directory")
    dir_contents = os.listdir(azure_directory)
    if not os.path.isfile(msal_file):
        raise Exception("msal_token_cache not found in the user's .azure directory")
    fp = open(msal_file, "rb")
    contents = fp.read()
    fp.close()
    if os_name == "nt":
        contents = decrypt_dpapi(contents).decode('utf-8')
    return json.loads(contents)



"""A module containing an Azure Devops HTTP Client"""
import logging
from azol.clients.oauth_http_client import OAuthHTTPClient
from azol.constants import OAuthResourceIDs
from datetime import datetime
from cryptography.hazmat.primitives.asymmetric import rsa
import base64

class AzureDevOpsClient( OAuthHTTPClient ):
    """
        An HTTP client for interacting with the Azure Devops API
    """

    def __init__( self, *args, **kwargs ):
        devops_resource_id=OAuthResourceIDs.AzureDevOps
        devops_base_url=f"https://dev.azure.com"

        super().__init__( oauth_resource=devops_resource_id,
                          base_url=devops_base_url, *args, **kwargs)

    def get_agent_pools( self, org_name ):
        '''
            Get all the agent pools that the current credential can read in the organization

            Args
                org_name - the name of the devops organization

            Returns:
                A list of agent pools
        '''
        query_params = { "api-version": "7.1-preview.1"}
        url = f"/{org_name}/_apis/distributedtask/pools"
        resp=self._send_request( url,  method="GET", query_parameters=query_params ) 
        response_data = resp.json()
        return response_data

    def get_agents( self, org_name, agent_pool_id ):
        '''
            Get all the agents in an agent pool

            Args
                org_name - the name of the devops organization

                agent_pool_id - the id of the agent pool

            Returns:
                A list of agents in the pool
        '''
        query_params = { 
            "api-version": "7.1-preview.1",
            "includeAssignedRequest": True,
            "includeCapabilities": True
        }
        url = f"{org_name}/_apis/distributedtask/pools/{agent_pool_id}/agents"
        resp=self._send_request( url,  method="GET", query_parameters=query_params ) 
        response_data = resp.json()
        return response_data

    def get_service_endpoint_types( self, org_name ):
        '''
            Get all the current service connection types

            Args
                org_name - the name of the devops organization

            Returns:
                A list of service connection objects
        '''
        query_params = { "api-version": "7.1-preview.1"}

        url = f"/{org_name}/_apis/serviceendpoint/types"
        resp=self._send_request( url,  method="GET", query_parameters=query_params ) 
        response_data = resp.json()
        return response_data

    def create_agent(self, org_name, pool_id, name):
        '''
            Create a new devops agent in the specified pool.

            Args:

                org_name - the name of the devops org that the pool is in

                pool_id - the id of the devops agent pool that the agent should be added to

                name - a unique name for the new devops agent

            Returns:
                A tuple ( agent_data, rsa_parameters )

                Where agent_data is information about the agent resulting from the HTTP response,
                and rsa_parameters is the base64 encoded rsa parameters used to create the agent.
        '''

        def get_b64encoded_int_bytes(integer):
            num_bytes = (integer.bit_length() + 7) // 8
            byte_representation = integer.to_bytes(num_bytes, byteorder='big')
            return base64.b64encode(byte_representation).decode()

        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
        )
        public_numbers = private_key.public_key().public_numbers()
        modulus=public_numbers.n
        exponent=public_numbers.e
        dq=private_key.private_numbers().dmq1
        dp=private_key.private_numbers().dmp1
        q=private_key.private_numbers().q
        p=private_key.private_numbers().p
        d=private_key.private_numbers().d
        inverseq=private_key.private_numbers().iqmp

        b64_encoded_modulus=get_b64encoded_int_bytes(modulus)
        b64_encoded_exponent=get_b64encoded_int_bytes(exponent)
        b64_encoded_dq = get_b64encoded_int_bytes(dq)
        b64_encoded_dp = get_b64encoded_int_bytes(dp)
        b64_encoded_q = get_b64encoded_int_bytes(q)
        b64_encoded_p = get_b64encoded_int_bytes(p)
        b64_encoded_d = get_b64encoded_int_bytes(d)
        b64_encoded_inverseq = get_b64encoded_int_bytes(inverseq)

        rsa_params = {
            "p": b64_encoded_p,
            "q": b64_encoded_q,
            "d": b64_encoded_d,
            "dp": b64_encoded_dp,
            "dq": b64_encoded_dq,
            "inverseQ": b64_encoded_inverseq,
            "modulus": b64_encoded_modulus,
            "exponent": b64_encoded_exponent
        }

        url = f"/{org_name}/_apis/distributedtask/pools/{pool_id}/agents"
        
        query_params = { "api-version": "7.1-preview.1"}
        body = {
            "systemCapabilities": {},
            "maxParallelism": 1,
            "authorization": {
                "publicKey": {
                    "exponent": b64_encoded_exponent,
                    "modulus": b64_encoded_modulus
                }
            },
            "name": name,
            "version": "4.251.0",
            "osDescription": "Microsoft Windows 10.0.26100",
            "status": 0
        }

        resp=self._send_request( url, json=body, method="POST", query_parameters=query_params ) 
        response_data = resp.json()
        return (response_data, rsa_params)

    def get_service_connection_role_assignments(self, org_name, project_id, endpoint_id):
        """
            Get all role assignments on a service connection(endpoint).

            Note that if the service endpoint/project id combination does not exist, this
            endpoint will still return 200 OK with an empty list as a result. Ensure that the
            correct project_id value is used.

            Args:
                org_name - The devops organization name

                project_id - The project ID that contains the service endpoint. Note that the
                             project name will not work, the name needs to be resolved to an ID
                
                endpoint_id - The id of the service endpoint
            
            Returns:
                A list of role assignments
        """
        query_params = { "api-version": "7.1"}
        url = (f"/{org_name}/_apis/securityroles/scopes/"
                f"distributedtask.serviceendpointrole/roleassignments/resources/{project_id}_{endpoint_id}" )
        raw_response = self._send_request( url, query_parameters=query_params)
        if raw_response.status_code != 200:
            logging.error( "ERROR while getting service conection role assignments" )
            logging.error( raw_response.content )
            raise Exception( "Could not fetch service connection role assignments" )
            
        return raw_response.json()["value"]

    def get_repositories( self, org_name, project_name ):
        '''
            Get all repositories in the project

            Args:
                org_name - The devops organization name

                project_name - The project name that contains the service endpoint

                Returns:
                    A list of repositories
        '''
        query_params = { "api-version": "7.1"}
        raw_response = self._send_request( f"/{org_name}/{project_name}/_apis/git/repositories",
                                           query_parameters=query_params )
        if raw_response.status_code != 200:
            logging.error( "ERROR while getting repositories" )
            logging.error( raw_response.content )
            raise Exception("Could not fetch repositories")
            
        return raw_response.json()

    def get_repository( self, org_name, project_name, repository_id ):
        '''
            Get a devOps repository

            Args:
                org_name - The devops organization name

                project_name - The project name that contains the service endpoint
                
                repository_id - The id of the repository
            
                Returns:
                    A python object containing information about the repository
        '''
        query_params = { "api-version": "7.1"}
        raw_response = self._send_request( f"/{org_name}/{project_name}/_apis/git/repositories/{repository_id}",
                                           query_parameters=query_params )
        if raw_response.status_code != 200:
            logging.error( "ERROR while getting repository" )
            logging.error( raw_response.content )
            raise Exception("Could not fetch repository")
            
        return raw_response.json()


    def get_pipelines( self, org_name, project_name ):
        '''
            Get all devOps pipelines

            Args:
                org_name - The devops organization name

                project_name - The project name that contains the service endpoint

                Returns:
                    A list of pipelines
        '''
        query_params = { "api-version": "7.1"}
        raw_response = self._send_request( f"/{org_name}/{project_name}/_apis/pipelines",
                                           query_parameters=query_params )
        if raw_response.status_code != 200:
            logging.error( "ERROR while getting pipelines" )
            logging.error( raw_response.content )
            raise Exception("Could not fetch pipelines")
            
        return raw_response.json()

    def get_pipeline( self, org_name, project_name, pipeline_id ):
        '''
            Get a devOps pipeline

            Args:
                org_name - The devops organization name

                project_name - The project name that contains the service endpoint
                
                pipeline_id - The id of the pipeline
            
                Returns:
                    A python object containing information about the pipeline
        '''
        query_params = { "api-version": "7.1-preview.1"}
        raw_response = self._send_request( f"/{org_name}/{project_name}/_apis/pipelines/{pipeline_id}",
                                           query_parameters=query_params )
        if raw_response.status_code != 200:
            logging.error( "ERROR while getting pipeline" )
            logging.error( raw_response.content )
            raise Exception("Could not fetch pipeline")
            
        return raw_response.json()

    def get_allowed_pipelines( self, org_name, project_name, endpoint_id ):
        """
            Get all pipelines that are authorized to use a service connection.

            This command returns an object that includes high level settings for the
            service connection as well as pipelines that are authorized.

            Args:
                org_name - The devops organization name

                project_name - The project name that contains the service endpoint
                
                endpoint_id - The id of the service endpoint
            
            Returns:
                An object containing the allowed pipelines, as high level settings for pipeline authorization
                for the service connection
        """
        query_params = { "api-version": "7.1-preview.1"}
        raw_response = self._send_request( f"/{org_name}/{project_name}/_apis/pipelines/pipelinepermissions/endpoint/{endpoint_id}", query_parameters=query_params )
        if raw_response.status_code != 200:
            logging.error( "ERROR while getting allowed pipelines for service connection" )
            logging.error( raw_response.content )
            raise Exception("Could not fetch allowed pipelines for service connection")
            
        return raw_response.json()

    def get_service_connection( self, org_name, project_name, endpoint_id ):
        """
            Get a service connections

            Note that this will still return 200 OK and an empty object if the endpoint or project
            does not exist. Ensure that correct values are used.

            Args:
                org_name - The devops organization name

                project_name - The project name that contains the service endpoint

                endpoint_id - The id of the service endpoint
            
            Returns:
                A service connection
        """
        query_params = { "api-version": "7.2-preview.4"}
        raw_response = self._send_request( f"/{org_name}/{project_name}/"
                                           f"_apis/serviceendpoint/endpoints/{endpoint_id}",
                                           query_parameters=query_params )
        if raw_response.status_code != 200:
            logging.error( "ERROR while getting devops project service connections" )
            logging.error( raw_response.json() )
            raise Exception("Could not fetch service connections for project")
        return raw_response.json()

    def try_download_permissions_report( self, report_url ):
        """
            Try to download a permissions report, or return None if it is not ready

            Args:
                report_url - the report url returned from create_permission_report

            returns:
                deserialized report object if successful, or None if the report is not
                ready yet
        """
        query_params = { "api-version": "7.2-preview.1"}
        r = raw_response = self._send_request( url=report_url,
                                          query_parameters=query_params )
        status = r.status_code
        if status == 404:
            error_type = r.json()["typeKey"]
            if error_type == "PermissionsReportDownloadNotAvailableException":
                logging.warning( "Permissions report is not ready for download" )
                return None
            else:
                logging.error( "ERROR while downloading permissions report" )
                logging.error( raw_response.json() )
                raise Exception("Could not download permissions report")
        elif status != 200:
            logging.error( "ERROR while downloading permissions report" )
            logging.error( raw_response.json() )
            raise Exception("Could not download permissions report")
        
        return r.json()

    def create_permission_report( self, org_name, resource_id, resource_type ):
        """
            Create a permissions report for the resource id and resource types.
            https://learn.microsoft.com/en-us/rest/api/azure/devops/permissionsreport/

            Example to create a report on project permissions: 
            client.create_permission_report("[org_Name]", 
                   "[projectId]", "ProjectGit")

            Args:
                org_name - The devops organization name

                resource_id - The Azure DevOps resource ID that should be reported on.
                              For example, the project ID, project collection ID, etc.

                resource_type - The type of the resource to fetch permissions for.
                                Alowed values: ProjectGit, ref, repo, release, tfvc
            
            Returns:
                A URL to download the permissions report
        """
        query_params = { "api-version": "7.2-preview.1"}
        body = {
            "descriptors": [],
            "reportName": str(datetime.now().isoformat()),
            "resources": [
                {
                    "resourceId": resource_id,
                    "resourceType": resource_type
                }
            ]
        }
        raw_response = self._send_request( f"/{org_name}/_apis/permissionsreport", 
                                          method="POST", json=body, query_parameters=query_params )
        if raw_response.status_code != 202:
            logging.error( "ERROR while creating devops permissions report" )
            logging.error( raw_response.json() )
            raise Exception("Could not create devops permissions report")
        return raw_response.json()["_downloadLink"]["href"]


    def get_service_connections( self, org_name, project_name ):
        """
            Get all service connections in a project

            Args:
                org_name - The devops organization name

                project_name - The project name that contains the service endpoint
            
            Returns:
                A list of service connection objects
        """
        query_params = { "api-version": "7.2-preview.4"}
        raw_response = self._send_request( f"/{org_name}/{project_name}/_apis/serviceendpoint/endpoints", query_parameters=query_params )
        if raw_response.status_code != 200:
            logging.error( "ERROR while getting devops project service connections" )
            logging.error( raw_response.json() )
            raise Exception("Could not fetch service connections for project")
        return raw_response.json()["value"]

    def get_all_service_connections( self, org_name ):
        """
            Get all service connections in an organization

            Args:
                org_name - The devops organization name

            Returns:
                A dictionary where the keys are project names, and the values
                are a list of service connections found in that project
        """
        projects = self.get_projects(org_name)
        results = {}

        for project in projects:
            results[project["name"]] = []
            service_connections = self.get_service_connections(org_name, project["name"])
            for sc in service_connections:
                results[project["name"]].append(sc)
        return results

    def get_projects( self, org_name ):
        """Get all projects that the current logged in credential has access to
        
            Args:
                org_name - The devops organization name

            Returns:
                A list of projects
        """
        query_params = { "api-version": "7.2-preview.4"}
        raw_response = self._send_request( f"/{org_name}/_apis/projects", query_parameters=query_params )

        if raw_response.status_code != 200:
            logging.error( "ERROR while getting devops projects" )
            logging.error( raw_response.json())
            raise Exception("Could not fetch Azure DevOps Projects")
        return raw_response.json()["value"]

    def get_connection_data( self, org_name ):
        """
            Get the currently logged on user's information from the /connectionData endpoint

            Args:
                org_name - The devops organization name

            Returns:
                an object containing user information
        """
        query_params = { "api-version": "7.2-preview.1"}
        raw_response = self._send_request(f"/{org_name}/_apis/connectionData",
                                          method="GET", query_parameters=query_params)
        if raw_response.status_code != 200:
            logging.error( "ERROR while getting logged in users connection data" )
            logging.error( raw_response.content)
            raise Exception("Could not get logged in user connection data from devops API")
        return raw_response.json()

    def get_self(self):
        """
            Get the currently logged on user's information from the /User endpoint

            Returns:
                an object containing user information
        """
        query_params = { "api-version": "7.2-preview.1"}
        raw_response = self._send_request(url="https://aex.dev.azure.com/_apis/User/User",
                                          method="GET", query_parameters=query_params)
        if raw_response.status_code != 200:
            logging.error( "ERROR while getting logged in user" )
            logging.error( raw_response.json())
            raise Exception("Could not get logged in user from devops API")
        return raw_response.json()
    
    def get_profile(self):
        """
            Get the currently logged on user's profiles from the /profiles/me endpoint

            Returns:
                a list of profile objects
        """
        query_params = { "api-version": "7.2-preview.1"}
        raw_response = self._send_request(url="https://app.vssps.visualstudio.com/_apis/profile/profiles/me",
                                          method="GET", query_parameters=query_params)
        if raw_response.status_code != 200:
            logging.error( "ERROR while getting logged in users profiles" )
            logging.error( raw_response.json())
            raise Exception("Could not get logged in users profiles from devops API")
        return raw_response.json()


    def get_organizations(self):
        """Get all organizations that the current logged in credential has access to
        
            Args:
                member_id - the id of the user, found from /profiles/me

            Returns:
                A list of organizations
        """
        profile_data = self.get_profile()

        query_params = { "api-version": "7.2-preview.1", "memberId": profile_data["id"]}
        raw_response = self._send_request( url=f"https://app.vssps.visualstudio.com/_apis/accounts",
                                          query_parameters=query_params )

        if raw_response.status_code != 200:
            logging.error( "ERROR while getting devops organizations" )
            logging.error( raw_response.json())
            raise Exception("Could not fetch Azure DevOps organizations")
        return raw_response.json()["value"]


"""Utilities to be run during Azure Devops Pipeline executions"""
from .azure_devops_utils import *

__all__ = [
    "is_self_hosted",
    "get_agent_home_directory",
    "is_windows",
    "get_rsa_credentials_file",
    "load_rsa_credentials",
    "get_agent_id",
    "get_agent_file",
    "get_credentials_file",
    "get_latest_session_id"
]



"""A module containing the parent credential class"""
import uuid
from azol.credentials.credential import Credential
from azol.models import DevOpsRSAParameters
import logging
import json
import base64

class DevOpsAgentCredential( Credential ):
    """
        Credential for authenticating as an Azure DevOps Agent

        The agent authenticates with an RSA key and a client_credential grant towards devops.
        The RSA key can be input as a devops agent credentials file path or as the raw RSA
        parameters, base64 encoded.
    """

    credentialType="devops_agent_credential"

    def __init__( self, rsa_parameters_file_path=None, rsa_parameters=None, *args, **kwargs ):
        super().__init__( *args, **kwargs )

        if rsa_parameters_file_path:
            if rsa_parameters:
                logging.warning( "DevOpsAgentCredential: rsa_parameters ignored because"
                                 "rsa_parameters_file_path argument was also supplied."
                                 "rsa_parameters_file_path argument will be used" )
            f=open(rsa_parameters_file_path, encoding="ascii")
            #print(f.readlines())
            params_json=json.load(f)
            f.close()

            b64_d = params_json["d"]
            b64_p = params_json["p"]
            b64_q = params_json["q"]
            b64_dq = params_json["dq"]
            b64_dp = params_json["dp"]
            b64_inverseq = params_json["inverseQ"]
            b64_modulus = params_json["modulus"]
            b64_exponent = params_json["exponent"]
            d = int.from_bytes(base64.b64decode(b64_d), "big")
            q = int.from_bytes(base64.b64decode(b64_q), "big")
            p = int.from_bytes(base64.b64decode(b64_p), "big")
            dq = int.from_bytes(base64.b64decode(b64_dq), "big")
            dp = int.from_bytes(base64.b64decode(b64_dp), "big")
            inverseq = int.from_bytes(base64.b64decode(b64_inverseq), "big")
            modulus = int.from_bytes(base64.b64decode(b64_modulus), "big")
            exponent = int.from_bytes(base64.b64decode(b64_exponent), "big")
            self.rsa_parameters = DevOpsRSAParameters(
                d=d,
                p=p,
                q=q,
                dp=dp,
                dq=dq,
                inverseq=inverseq,
                modulus=modulus,
                exponent=exponent
            )
        elif rsa_parameters:
            if type(rsa_parameters) == "DevOpsRSAParameters":
                self.rsa_parameters = rsa_parameters
            else:
                b64_d = rsa_parameters["d"]
                b64_p = rsa_parameters["p"]
                b64_q = rsa_parameters["q"]
                b64_dq = rsa_parameters["dq"]
                b64_dp = rsa_parameters["dp"]
                b64_inverseq = rsa_parameters["inverseQ"]
                b64_modulus = rsa_parameters["modulus"]
                b64_exponent = rsa_parameters["exponent"]
                d = int.from_bytes(base64.b64decode(b64_d), "big")
                q = int.from_bytes(base64.b64decode(b64_q), "big")
                p = int.from_bytes(base64.b64decode(b64_p), "big")
                dq = int.from_bytes(base64.b64decode(b64_dq), "big")
                dp = int.from_bytes(base64.b64decode(b64_dp), "big")
                inverseq = int.from_bytes(base64.b64decode(b64_inverseq), "big")
                modulus = int.from_bytes(base64.b64decode(b64_modulus), "big")
                exponent = int.from_bytes(base64.b64decode(b64_exponent), "big")
                self.rsa_parameters = DevOpsRSAParameters(
                    d=d,
                    p=p,
                    q=q,
                    dp=dp,
                    dq=dq,
                    inverseq=inverseq,
                    modulus=modulus,
                    exponent=exponent
                )


"""Module containing utilities for working with Azure DevOps Agents. For use in Azure Devops pipeline tasks."""

import os
from azol.utils.local.windows import decrypt_dpapi
import json
from azol.models.devops_rsa_parameters import DevOpsRSAParameters
import base64
import re

def is_self_hosted():
    '''
        For use in Azure Devops pipeline tasks.

        Tries to identify if this is a self-hosted agent or a Microsoft-hosted agent. 
    '''
    self_hosted = os.getenv("AGENT_ISSELFHOSTED")
    if self_hosted == "0":
        return False
    if self_hosted == "1":
        return True
    return None

def get_agent_home_directory():
    '''
        For use in Azure Devops pipeline tasks.

        Tries to read the agent home directory environment variable
    '''
    return os.getenv("AGENT_HOMEDIRECTORY")

def is_windows():
    '''
        For use in Azure Devops pipeline tasks.

        Returns true if windows OS else false
    '''
    if os.name == "nt":
        return True
    else:
        return False

def get_rsa_credentials_file():
    '''
        For use in Azure Devops pipeline tasks.

        Search and read the agent's .credentials_rsaparams file. Decrypt it if
        this is a Windows machine.

        returns a dictionary containing the raw contents of the file.
    '''


    home_directory = get_agent_home_directory()
    if is_self_hosted():
        rsa_credentials_file = os.path.normpath( home_directory + "/.credentials_rsaparams") 
        raw_contents = b""
        with open(rsa_credentials_file, "rb") as f:
            raw_contents = raw_contents + f.read()
        if is_windows():
            decrypted_bytes = decrypt_dpapi(raw_contents).decode('utf-8')
        else:
            decrypted_bytes = raw_contents
            
        credentials = json.loads(decrypted_bytes)
        return credentials
    else:
        raise Exception("Microsoft hosted agents do not have a .credentuals_rsaparams file")

def load_rsa_credentials(credentials_dict):
    '''
        For use in Azure Devops pipeline tasks.

        Load a .credentials_rsaparams file into a DevOpsRSAParameters object

            Args:
                credentials_dict - (dict) A dictionary containing the deserialized contents 
                                        of a -credentials_rsaparams file
            Returns:
                A DevOpsRSAParameters object containing the RSA parameters

    '''
    b64_d = credentials_dict["d"]
    b64_p = credentials_dict["p"]
    b64_q = credentials_dict["q"]
    b64_dq = credentials_dict["dq"]
    b64_dp = credentials_dict["dp"]
    b64_inverseq = credentials_dict["inverseQ"]
    b64_modulus = credentials_dict["modulus"]
    b64_exponent = credentials_dict["exponent"]
    d = int.from_bytes(base64.b64decode(b64_d), "big")
    q = int.from_bytes(base64.b64decode(b64_q), "big")
    p = int.from_bytes(base64.b64decode(b64_p), "big")
    dq = int.from_bytes(base64.b64decode(b64_dq), "big")
    dp = int.from_bytes(base64.b64decode(b64_dp), "big")
    inverseq = int.from_bytes(base64.b64decode(b64_inverseq), "big")
    modulus = int.from_bytes(base64.b64decode(b64_modulus), "big")
    exponent = int.from_bytes(base64.b64decode(b64_exponent), "big")
    rsa_parameters = DevOpsRSAParameters(
        d=d,
        p=p,
        q=q,
        dp=dp,
        dq=dq,
        inverseq=inverseq,
        modulus=modulus,
        exponent=exponent
    )
    return rsa_parameters


def get_agent_id():
    '''
        For use in Azure Devops pipeline tasks.

        return the agent ID from environment variables

         Returns:
            The agent ID, or None if the environment variable cannot be found

    '''
    return os.getenv("AGENT_ID")


def get_agent_file():
    '''
        For use in Azure Devops pipeline tasks.

        Get the contents of the agent file for the build agent

         Returns:
            A dictionary containing the contents of the .agent file
    '''
    home_directory = get_agent_home_directory()
    agent_file = os.path.normpath( home_directory + "/.agent") 
    f = open(agent_file, "rb")
    agent_data=json.load(f)
    f.close()
    return agent_data

def get_credentials_file():
    '''
        For use in Azure Devops pipeline tasks.

        Get the contents of the .credentials file for the build agent

         Returns:
            A dictionary containing the contents of the .credentials file
    '''
    home_directory = get_agent_home_directory()
    credentials_file = os.path.normpath( home_directory + "/.credentials") 
    f = open(credentials_file, "rb")
    agent_data=json.load(f)
    f.close()
    return agent_data

def get_latest_session_id():
    '''
        For use in Azure Devops pipeline tasks.

        Try to get the latest session ID from the agent logs.

        The session ID is needed in order to kill an agent and spoof it. 
        The session ID can be found in 'agent_home/_diag/Agent_date-utc.logs' files.
        This method searches for the most recent file, and looks for the last session
        ID in those logs. This shoud represent the current session ID

         Returns:
            A session ID if one was found, else None
    '''
    home_directory = get_agent_home_directory()
    diagnostics_directory = os.path.normpath(home_directory + "/_diag") 
    files = [f for f in os.listdir(diagnostics_directory) 
             if os.path.isfile(os.path.join(diagnostics_directory, f))
             and f.startswith("Agent")]
    files.sort()
    latest = files[-1]
    full_path=os.path.join(diagnostics_directory, latest)
    with open(full_path) as f: content = f.read()
    q = re.compile("[sS]ession '([a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12})'",re.S|re.M)
    session_ids = q.findall(content)
    if session_ids == []: return None
    return session_ids[-1]
